# -*- coding: utf-8 -*-

"""
jsongit.api
"""

import os
import pygit2

from .models import Repository
import utils

def init(path=None, repo=None, **kwargs):
    """Obtain a :class:`Repository`.  Either a path to a repo or an
    existing repo must be provided.

    :param path:
        The path to a repository. If it is a path that does not exist, a new
        bare git repository will be initialized there.  If it is a path that
        does exist, then the directory will be used as a repository.
    :type path: string
    :param repo: An existing repository object.
    :type repo: :class:`pygit2.Repository`
    :param dumps:
        (optional) An alternate function to use when dumping data.  Defaults
        to :func:`json.dumps`.
    :type dumps: func
    :param loads:
        (optional) An alternate function to use when loading data.  Defaults
        to :func:`json.loads`.
    :type loads: func

    :returns: A repository reference
    :rtype: :class:`Repository`
    """
    if repo and path:
        raise TypeError("Cannot define repo and path")
    if path:
        if os.path.isdir(path):
            repo = pygit2.Repository(path)
        else:
            bare = kwargs.pop('bare', True) # bare repo by default
            repo = pygit2.init_repository(path, bare)
    if not repo:
        raise TypeError("Missing repo or path")
    dumps = kwargs.pop('dumps', utils.import_json().dumps)
    loads = kwargs.pop('loads', utils.import_json().loads)
    return Repository(repo, dumps, loads)



# -*- coding: utf-8 -*-

"""
jsongit.author
"""

from time import altzone, daylight, timezone
from time import time as curtime
from pygit2 import Signature
import subprocess

from .exceptions import NoGlobalSettingError

def global_config(name):
    """Find a git --global setting

    :param name: the name of the setting
    :type name: string
    :return: the value of the setting
    :rtype: string
    :raises: NoGlobalSettingError
    """
    # TODO libgit2 provides an interface for this, but pygit2 does not.  Should
    # patch pygit2 to provide it.  In the interim, we must use subprocess.
    popen = subprocess.Popen(['git', 'config', '--global', name],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    out, err = popen.communicate()
    if popen.returncode == 0:
        return out.rstrip()
    else:
        raise NoGlobalSettingError(name)

def signature(name, email, time=None, offset=None):
    """Convenience method to generate pygit2 signatures.

    :param name: The name in the signature.
    :type name: string
    :param email: The email in the signature.
    :type email: string
    :param time: (optional) the time for the signature, in UTC seconds.
        Defaults to current time.
    :type time: int
    :param offset:
        (optional) the time offset for the signature, in minutes.
        Defaults to the system offset.
    :type offset: int

    :returns: a signature
    :rtype: pygit2.Signature
    """
    offset = offset or (altzone / 60 if daylight else timezone / 60)
    time = time or int(curtime())
    return Signature(name, email, time, offset)

def import_json():
    try:
        import simplejson
        return simplejson
    except ImportError:
        import json
        return json




# -*- coding: utf-8 -*-

from jsongit.models import Diff, Conflict
import helpers
import itertools


class DiffTest(helpers.unittest.TestCase):

    @property
    def fixtures(self):
        return [ 'foo', 7,
                ['foo', 'bar', 'baz'],
                {'roses': 'red', 'violets':'blue'}]

    def test_diff_different_types_favor_second(self):
        for a, b in itertools.permutations(self.fixtures, 2):
            diff = Diff(a, b)
            self.assertEquals(b, diff.replace)
            self.assertEquals(b, diff.apply(a))

    def test_no_changes(self):
        for f in self.fixtures:
            diff = Diff(f, f)
            self.assertIsNone(diff.replace)
            self.assertIsNone(diff.update)
            self.assertIsNone(diff.remove)
            self.assertIsNone(diff.append)
            self.assertEquals(f, diff.apply(f))

    def test_diff_numbers(self):
        a = 4
        b = 7
        diff = Diff(a, b)
        self.assertEquals(7, diff.replace)
        self.assertEquals(b, diff.apply(a))

    def test_diff_strings(self):
        a = 'foo'
        b = 'bar'
        diff = Diff(a, b)
        self.assertEquals('bar', diff.replace)
        self.assertEquals(b, diff.apply(a))

    def test_diff_array_append(self):
        a = ['foo']
        b = ['foo', 'bar']
        diff = Diff(a, b)
        self.assertEquals({1: 'bar'}, diff.append)
        self.assertEquals(b, diff.apply(a))

    def test_diff_array_update(self):
        a = ['foo', 'bar']
        b = ['foo', 'baz']
        diff = Diff(a, b)
        self.assertEquals({1: 'baz'}, diff.update)
        self.assertEquals(b, diff.apply(a))

    def test_diff_array_remove(self):
        a = ['foo', 'bar']
        b = ['foo']
        diff = Diff(a, b)
        self.assertEquals({1: 'bar'}, diff.remove)
        self.assertEquals(b, diff.apply(a))

    def test_diff_array_nested_append(self):
        a = ['foo', ['bar']]
        b = ['foo', ['bar', 'baz']]
        diff = Diff(a, b)
        self.assertEquals({1: 'baz'}, diff.update[1].append)
        self.assertEquals(b, diff.apply(a))

    def test_diff_array_nested_update(self):
        a = ['foo', ['bar', 'baz']]
        b = ['foo', ['bar', 'bazzz']]
        diff = Diff(a, b)
        self.assertEquals({1: 'bazzz'}, diff.update[1].update)
        self.assertEquals(b, diff.apply(a))

    def test_diff_array_nested_remove(self):
        a = ['foo', ['bar', 'baz']]
        b = ['foo', ['bar']]
        diff = Diff(a, b)
        self.assertEquals({1: 'baz'}, diff.update[1].remove)
        self.assertEquals(b, diff.apply(a))

    def test_diff_empty_dict(self):
        """json_diff bugs out on empty dicts normally.
        """
        a = {}
        b = {'key': 'value'}
        diff = Diff(a, b)
        self.assertEquals({'key': 'value'}, diff.append)
        self.assertEquals(b, diff.apply(a))

    def test_diff_dict_append(self):
        a = {'roses': 'red'}
        b = {'roses': 'red', 'violets': 'blue'}
        diff = Diff(a, b)
        self.assertEquals({'violets': 'blue'}, diff.append)
        self.assertEquals(b, diff.apply(a))

    def test_diff_dict_update(self):
        a = {'roses': 'red', 'violets': 'blue'}
        b = {'roses': 'red', 'violets': 'aqua'}
        diff = Diff(a, b)
        self.assertEquals({'violets': 'aqua'}, diff.update)
        self.assertEquals(b, diff.apply(a))

    def test_diff_dict_remove(self):
        a = {'roses': 'red', 'violets': 'blue'}
        b = {'roses': 'red'}
        diff = Diff(a, b)
        self.assertEquals({'violets': 'blue'}, diff.remove)
        self.assertEquals(b, diff.apply(a))

    def test_diff_dict_nested_append(self):
        a = {'flowers': {'roses': 'red'}}
        b = {'flowers': {'roses': 'red', 'violets': 'blue'}}
        diff = Diff(a, b)
        self.assertEquals({'violets': 'blue'}, diff.update['flowers'].append)
        self.assertEquals(b, diff.apply(a))

    def test_diff_dict_nested_update(self):
        a = {'flowers': {'roses': 'red', 'violets': 'blue'}}
        b = {'flowers': {'roses': 'red', 'violets': 'aqua'}}
        diff = Diff(a, b)
        self.assertEquals({'violets': 'aqua'}, diff.update['flowers'].update)
        self.assertEquals(b, diff.apply(a))

    def test_diff_dict_nested_remove(self):
        a = {'flowers': {'roses': 'red', 'violets': 'blue'}}
        b = {'flowers': {'roses': 'red'}}
        diff = Diff(a, b)
        self.assertEquals({'violets': 'blue'}, diff.update['flowers'].remove)
        self.assertEquals(b, diff.apply(a))

    def xtest_diff_scalar_replace_no_conflict(self):
        a = 'foo'
        b = 'bar'
        c = 'bar'
        conflict = Conflict(Diff(a, b), Diff(a, c))
        self.assertFalse(conflict)

    def test_diff_scalar_replace_conflict(self):
        a = 'foo'
        b = 'bar'
        c = 'baz'
        conflict = Conflict(Diff(a, b), Diff(a, c))
        self.assertEquals(('bar', 'baz'), conflict.replace)

    def test_diff_array_append_conflict(self):
        a = ['foo']
        b = ['foo', 'bar']
        c = ['foo', 'baz']
        conflict = Conflict(Diff(a, b), Diff(a, c))
        # self.assertEquals(('bar', 'baz'), conflict[1].append)
        self.assertEquals({1: ('bar', 'baz')}, conflict.append)

    def test_diff_array_update_conflict(self):
        a = ['foo', 'bar']
        b = ['foo', 'baz']
        c = ['foo', 'bazzz']
        conflict = Conflict(Diff(a, b), Diff(a, c))
        # self.assertEquals(('baz', 'bazzz'), conflict[1].update)
        self.assertEquals({1: ('baz', 'bazzz')}, conflict.update)

    def test_diff_array_remove_conflict(self):
        a = ['foo', 'bar']
        b = ['foo', 'baz']
        c = ['foo']
        conflict = Conflict(Diff(a, b), Diff(a, c))
        # self.assertEquals(('baz', None), conflict[1].update)
        # self.assertEquals((None, 'bar'), conflict[1].remove)
        self.assertEquals({1: ('baz', None)}, conflict.update)
        self.assertEquals({1: (None, 'bar')}, conflict.remove)

    def xtest_diff_array_nested_append_conflict(self):
        a = ['boo', ['foo']]
        b = ['boo', ['foo', 'bar']]
        c = ['boo', ['foo', 'baz']]
        conflict = Conflict(Diff(a, b), Diff(a, c))
        self.assertEquals({1: ('bar', 'baz')}, conflict.update[1].append)

    def xtest_diff_array_nested_update_conflict(self):
        a = ['boo', ['foo', 'bar']]
        b = ['boo', ['foo', 'baz']]
        c = ['boo', ['foo', 'bazzz']]
        conflict = Conflict(Diff(a, b), Diff(a, c))
        # self.assertEquals(('baz', 'bazzz'), conflict[1].update)
        self.assertEquals({1: ('baz', 'bazzz')}, conflict.update)

    def xtest_diff_array_nested_remove_conflict(self):
        a = ['boo', ['foo', 'bar']]
        b = ['boo', ['foo', 'baz']]
        c = ['boo', ['foo']]
        conflict = Conflict(Diff(a, b), Diff(a, c))
        # self.assertEquals(('baz', None), conflict[1].update)
        # self.assertEquals((None, 'bar'), conflict[1].remove)
        self.assertEquals({1: ('baz', None)}, conflict.update)
        self.assertEquals({1: (None, 'bar')}, conflict.remove)

    def test_diff_dict_append_conflict(self):
        a = {'roses': 'red'}
        b = {'roses': 'red', 'violets': 'blue'}
        c = {'roses': 'red', 'violets': 'magenta'}
        conflict = Conflict(Diff(a, b), Diff(a, c))
        self.assertEquals({'violets': ('blue', 'magenta')}, conflict.append)

    def test_diff_dict_update_conflict(self):
        a = {'roses': 'red'}
        b = {'roses': 'blue'}
        c = {'roses': 'white'}
        conflict = Conflict(Diff(a, b), Diff(a, c))
        self.assertEquals({'roses': ('blue', 'white')}, conflict.update)

    def test_diff_dict_remove_conflict(self):
        a = {'roses': 'red', 'violets':'blue'}
        b = {'roses': 'red', 'violets': 'magenta'}
        c = {'roses': 'red'}
        conflict = Conflict(Diff(a, b), Diff(a, c))
        self.assertEquals({'violets': ('magenta', None)}, conflict.update)
        self.assertEquals({'violets': (None, 'blue')}, conflict.remove)




from helpers import RepoTestCase

import jsongit

class TestLog(RepoTestCase):

    def test_no_log(self):
        """Starts with just the most recent step.
        """
        self.repo.commit('fresh', {'foo': 'bar'})
        gen = self.repo.log('fresh')
        commit = gen.next()
        self.assertEquals(commit, self.repo.get('fresh').head)
        self.assertEquals(commit.object.value, self.repo.get('fresh').value)

    def test_default_log(self):
        """Should step through commits backwards all the way by default.
        """
        self.repo.commit('foo', 'step 1')
        self.repo.commit('foo', 'step 2')
        self.repo.commit('foo', 'step 3')
        self.repo.commit('foo', 'step 4')

        gen = self.repo.log('foo')

        self.assertEquals('step 4', gen.next().object.value)
        self.assertEquals('step 3', gen.next().object.value)
        self.assertEquals('step 2', gen.next().object.value)
        self.assertEquals('step 1', gen.next().object.value)
        with self.assertRaises(StopIteration):
            gen.next()

    def test_reverse_log(self):
        """Can specify a reverse log.
        """
        self.repo.commit('foo', 'step 1')
        self.repo.commit('foo', 'step 2')
        self.repo.commit('foo', 'step 3')
        self.repo.commit('foo', 'step 4')

        gen = self.repo.log('foo', order=jsongit.GIT_SORT_REVERSE)

        self.assertEquals('step 1', gen.next().object.value)
        self.assertEquals('step 2', gen.next().object.value)
        self.assertEquals('step 3', gen.next().object.value)
        self.assertEquals('step 4', gen.next().object.value)
        with self.assertRaises(StopIteration):
            gen.next()

    def test_log_messages(self):
        """Retrieves log messages.
        """
        self.repo.commit('foo', 'bar', message="I think this is right.")
        self.repo.commit('foo', 'baz', message="Nope, it was wrong.")

        gen = self.repo.log('foo')

        self.assertEquals("Nope, it was wrong.", gen.next().message)
        self.assertEquals("I think this is right.", gen.next().message)

    def test_log_authors(self):
        """Retrieves authors.
        """
        bob = jsongit.utils.signature('bob', 'bob@bob.com')
        dan = jsongit.utils.signature('dan', 'dan@dan.com')
        self.repo.commit('foo', 'bar', author=bob)
        self.repo.commit('foo', 'baz', author=dan)

        gen = self.repo.log('foo')

        c = gen.next()
        self.assertEquals("dan", c.author.name)
        self.assertEquals("dan@dan.com", c.author.email)
        c = gen.next()
        self.assertEquals("bob", c.author.name)
        self.assertEquals("bob@bob.com", c.author.email)

    def test_merge_log(self):
        """Provides history for merged data that includes all parents.

        Changes should be registered at point they are merged.
        """
        self.repo.commit('foo', {'roses': 'red'})
        self.repo.merge('bar', 'foo')
        self.repo.commit('foo', {'roses': 'red', 'violets': 'blue'})
        self.repo.commit('bar', {'roses': 'red', 'lilacs': 'purple'})
        merge = self.repo.merge('bar', 'foo')
        self.assertTrue(merge.success)

        gen = self.repo.log('bar')

        self.assertEquals({'roses': 'red', 'violets':'blue', 'lilacs':'purple'},
                          gen.next().object.value)
        self.assertEquals({'roses': 'red', 'violets':'blue'},
                          gen.next().object.value,
                         "Merge occurred after modification to bar.")
        self.assertEquals({'roses': 'red', 'lilacs':'purple'},
                          gen.next().object.value)
        self.assertEquals({'roses': 'red'}, gen.next().object.value)
        self.assertEquals({'roses': 'red'}, gen.next().object.value)
        with self.assertRaises(StopIteration):
            gen.next()




import sys
import helpers
import threading
import random as global_random
import string

FEW = 3
LOTS = 10
MAX_DEPTH = 1

def random_number(r, j, l, min=-sys.maxint, max=sys.maxint):
    r.jumpahead(j + l)
    return r.random() * (max - min) + min

def random_string(r, j, l, min=1, max=50):
    r.jumpahead(j + l)
    return ''.join(r.choice(string.ascii_letters +
                            string.digits)
                   for i in xrange(r.randint(min, max)))

def random_array(r, j, l, min=1, max=10):
    if l > MAX_DEPTH:
        return []
    r.jumpahead(j + l)
    return [random_object(r, i, l + 1) for i in xrange(r.randint(min, max))]

def random_dict(r, j, l, min=1, max=10):
    if l > MAX_DEPTH:
        return {}
    r.jumpahead(j + l)
    return dict([(r.choice([random_string, random_number])(r, i, l),
                  random_object(r, i, l + 1))
                 for i in xrange(r.randint(min, max))])

def random_object(r, j, l):
    r.jumpahead(j + l)
    return r.choice([random_array, random_dict, random_string, random_number])(r, j, l)

def commit(repo, num_commits):
    r = global_random.Random()
    r.seed()
    thread_id = threading.current_thread().ident
    for i in xrange(num_commits):
        repo.commit(random_string(r, thread_id, 0), random_object(r, thread_id, 0))

class TestRepoThreading(helpers.RepoTestCase):

    def do_with_threads(self, size, func, *args):
        """Generate and start size threads with target func.
        """
        pool = [threading.Thread(target=func, args=args) for i in range(size)]
        for thread in pool:
            thread.start()
        return pool

    def join_threads(self, pool):
        for thread in pool:
            thread.join()

    def test_commit_once_few(self):
        """Make one commit each.
        """
        pool = self.do_with_threads(FEW, commit, self.repo, 1)
        self.join_threads(pool)

    def test_commit_once_lots(self):
        """Make one commit each.
        """
        pool = self.do_with_threads(LOTS, commit, self.repo, 1)
        self.join_threads(pool)

    def test_commit_lots_few(self):
        """Make several commits each.
        """
        pool = self.do_with_threads(FEW, commit, self.repo, 10)
        self.join_threads(pool)

    def test_commit_lots_lots(self):
        """Make several commits each.
        """
        pool = self.do_with_threads(LOTS, commit, self.repo, 10)
        self.join_threads(pool)







import jsongit
import helpers
import os
import json
import pygit2
import shutil

class TestJsonGitRepository(helpers.RepoTestCase):

    def test_create_path_arg(self):
        repo = jsongit.init('test_create_repo')
        self.assertTrue(os.path.isdir('test_create_repo'))
        repo.destroy()

    def test_create_path_kwarg(self):
        repo = jsongit.init(path='test_create_repo_kwarg')
        self.assertTrue(os.path.isdir('test_create_repo_kwarg'))
        repo.destroy()

    def test_destroy_repo(self):
        repo = jsongit.init('test_destroy_repo')
        repo.destroy()
        self.assertFalse(os.path.isdir('test_destroy_repo'))

    def test_has(self):
        """
        Test existence of extant key
        """
        self.repo.commit('foo', {'roses': 'red'})
        self.assertTrue(self.repo.has('foo'))

    def test_does_not_have(self):
        """
        Test existence of nonexistent key
        """
        self.assertFalse(self.repo.has('foo'))

    def test_get_nonexistent(self):
        """
        Arbitrary path should cause KeyError
        """
        with self.assertRaises(KeyError):
            self.assertEqual({}, self.repo.get('nuthin'))

    def test_commit_number(self):
        """Support numbers.
        """
        self.repo.commit('number', 7)
        self.assertEqual(7, self.repo.get('number').value)

    def test_commit_string(self):
        """Support strings
        """
        self.repo.commit('string', 'foo bar baz')
        self.assertEqual('foo bar baz', self.repo.get('string').value)

    def test_commit_list(self):
        """Support lists
        """
        self.repo.commit('list', ['foo', 'bar', 'baz'])
        self.assertEqual(['foo', 'bar', 'baz'], self.repo.get('list').value)

    def test_commit_dict(self):
        """Support dicts
        """
        self.repo.commit('dict', {'foo': 'bar'})
        self.assertEqual({'foo': 'bar'}, self.repo.get('dict').value)

    def test_commit_info(self):
        """Get head and info
        """
        self.repo.commit('obj', {'foo': 'bar'},
                         message='my special message',
                         author=jsongit.utils.signature('sally', 's@s.com'))
        obj = self.repo.get('obj')
        self.assertEquals('my special message', obj.head.message)
        self.assertEquals('sally', obj.head.author.name)
        self.assertEquals('s@s.com', obj.head.author.email)

    def test_get_back(self):
        """Get historical head and info
        """
        self.repo.commit('obj', {'foo': 'bar'},
                         message='my special message',
                         author=jsongit.utils.signature('sally', 's@s.com'))
        self.repo.commit('obj', {'bar': 'baz'})
        obj = self.repo.get('obj', back=1)
        self.assertEquals({'foo': 'bar'}, obj.value)
        self.assertEquals('my special message', obj.head.message)
        self.assertEquals('sally', obj.head.author.name)
        self.assertEquals('s@s.com', obj.head.author.email)

    def test_head_back_too_far(self):
        """Should get IndexError if we try to go back too far.
        """
        self.repo.commit('obj', {'foo': 'bar'})
        self.repo.commit('obj', {'bar': 'baz'})
        with self.assertRaises(IndexError):
            self.repo.get('obj', back=2)

    def test_copy(self):
        """ Copy an existing object using merge.
        """
        self.repo.commit('foo', {'roses': 'red'})
        self.repo.merge('bar', 'foo')
        self.assertEqual({'roses': 'red'}, self.repo.get('bar'))

    def test_merge_commit(self):
        """Merge explicitly to a commit.
        """
        self.repo.commit('foo', {'roses': 'red'})
        self.repo.merge('bar', commit=self.repo.get('foo').head)
        self.assertEqual({'roses': 'red'}, self.repo.get('bar'))

    def test_merge_old(self):
        """Fast forward explicitly to an older commit.
        """
        self.repo.commit('foo', {'roses': 'red'})
        self.repo.commit('foo', {'violets': 'blue'})
        self.repo.merge('bar', commit=self.repo.get('foo', back=1).head)
        self.assertEqual({'roses': 'red'}, self.repo.get('bar'))

    def test_merge_already_existing(self):
        """
        Merge should not overwrite independent already existing value.
        """
        self.repo.commit('foo', {'roses': 'red'})
        self.repo.commit('bar', {'violets': 'blue'})
        self.repo.merge('bar', 'foo')
        self.assertEqual({'violets': 'blue'}, self.repo.get('bar'))

    def test_merge_self(self):
        """
        Merging identical keys should raise an error.
        """
        self.repo.commit('foo', {'roses': 'red'})
        with self.assertRaises(ValueError):
            self.repo.merge('foo', 'foo')

    def test_commit_updating(self):
        """
        Can use commit to update.
        """
        self.repo.commit('foo', {'roses':'red'})
        self.repo.commit('foo', {})
        self.assertEqual({}, self.repo.get('foo'))

    def test_not_json(self):
        """
        Cannot commit something that cannot be converted to JSON to the db.
        """
        not_json = [lambda x: x]
        for item in not_json:
            with self.assertRaises(jsongit.NotJsonError):
                self.repo.commit('foo', item)

    def test_key_must_be_string(self):
        """
        Cannot commit to a non-string key in the repo.
        """
        not_strings = [lambda x: x, 4, None, ['foo', 'bar'], {'foo': 'bar'}]
        for item in not_strings:
            with self.assertRaises(jsongit.InvalidKeyError):
                self.repo.commit(item, {'foo': 'bar'})

    def test_all_keys_in_master(self):
        """All keys should be real blobs in the repo's head.
        """
        self.repo.commit('roses', 'red')
        self.repo.commit('violets', 'blue')

        pygit2_repo = self.repo._repo
        head_ref = pygit2_repo.lookup_reference('HEAD').resolve()
        head_commit = pygit2_repo[head_ref.oid]
        tree = head_commit.tree
        self.assertIn('roses', tree)
        self.assertIn('violets', tree)
        self.assertEquals(json.dumps('red'), tree['roses'].to_object().data)
        self.assertEquals(json.dumps('blue'), tree['violets'].to_object().data)

    def test_overlapping_paths(self):
        """Should throw error if key overlaps with directory.
        """
        self.repo.commit('path/to', 'foo')
        with self.assertRaises(jsongit.InvalidKeyError):
            self.repo.commit('path/to/key', 'bar')

    def test_optional_absolute(self):
        """Absolute-ish path should make no difference.
        """
        self.repo.commit('/absolute', 'foo')
        self.assertEquals('foo', self.repo.get('/absolute').value)
        self.assertEquals('foo', self.repo.get('absolute').value)

    def test_no_directory_path(self):
        """Should not be able to store data in a directory.
        """
        with self.assertRaises(jsongit.InvalidKeyError):
            self.repo.commit('path/to/directory/', 'foo')

    def xtest_nonbare_repo_master_tree(self):
        """Should be able to work with an existing non-bare repo.  Commits
        should appear as standard blobs in the master tree.
        """
        PATH = 'nonbare'
        if os.path.lexists(PATH):
            self.fail("Cannot run test, something exists at %s" % PATH)

        try:
            nonbare = pygit2.init_repository(PATH, False)

            repo = jsongit.init(repo=nonbare)
            repo.commit('foo', 'bar')
            repo.commit('path/to/file.json', {'roses': 'red', 'violets': 'blue'})

            # equivalent to checkout master
            head = nonbare.lookup_reference('HEAD').resolve()
            # index = nonbare.index
            # index.read_tree(nonbare[head.oid].tree.oid)
            # index.write()
        finally:
            shutil.rmtree(PATH)



# -*- coding: utf-8 -*-

"""
jsongit.wrappers

These classes provide limited interfaces to pygit2 and json_diff constructs.
"""

import json_diff
import itertools
import copy

class Commit(object):
    """A wrapper around :class:`pygit2.Commit` that links to a specific key
    in the repo.
    """

    def __init__(self, repo, pygit2_commit):
        self._repo = repo
        self._key = pygit2_commit.tree[0].name
        self._commit = pygit2_commit

    def __eq__(self, other):
        return self.oid == other.oid

    def __str__(self):
        return self._commit.__str__()

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, self._commit.__repr__())

    @property
    def object(self):
        """
        :returns: the data object associated with this commit.
        :rtype: :class:`Object`
        """
        return self._repo.get(commit=self)

    @property
    def key(self):
        """
        :returns: the key associated with this commit.
        :rtype: string
        """
        return self._key

    @property
    def oid(self):
        """
        :returns: The unique 20-byte ID of this Commit.
        :rtype: string
        """
        return self._commit.oid

    @property
    def hex(self):
        """
        :returns: The unique 40-character hex representation of this commit's ID.
        :rtype: string
        """
        return self._commit.hex

    @property
    def message(self):
        """
        :returns: The message associated with this commit.
        :rtype: string
        """
        return self._commit.message

    @property
    def author(self):
        """
        :returns: The author of this commit.
        :rtype: :class:`pygit2.Signature`
        """
        return self._commit.author

    # @property
    # def parents(self):
    #     """
    #     :returns: the parents of this commit.
    #     :rtype: list of :class:`Object`
    #     """
    #     #TODO might generate commits with the wrong key?
    #     return [Commit(self._repo, self.key, c) for c in self._commit.parents]


class DiffWrapper(object):
    """An internal wrapper for :mod:`json_diff`.
    """

    def __init__(self, diff):
        if Diff.is_json_diff(diff):
            # wrap recursive updates
            if Diff.UPDATE in diff:
                update = diff[Diff.UPDATE]
                for k, v in update.iteritems():
                    update[k] = DiffWrapper(v)
            self._replace = None
        else:
            self._replace = diff
            diff = {} if diff is None else diff

        self._diff = diff

    def __str__(self):
        return self._diff.__str__()

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, self._diff.__repr__())

    def __getitem__(self, k):
        return self._diff[k]

    def __eq__(self, other):
        return self._diff == other

    @property
    def remove(self):
        """A dict of removed keys and their values.
        """
        return self._diff.get(Diff.REMOVE)

    @property
    def update(self):
        """A DiffWrapper
        """
        return self._diff.get(Diff.UPDATE)

    @property
    def append(self):
        """A dict of appended keys and their values.
        """
        return self._diff.get(Diff.APPEND)

    @property
    def replace(self):
        """The diff is simply to replace wholesale.
        """
        return self._replace

    def apply(self, original):
        """Return an object modified with the changes in this diff.

        :param original: the object to apply the diff to.
        :type original: list, dict, number, or string

        :returns: the modified object
        :rtype: list, dict, number, or string
        """
        if self.replace:
            return self.replace
        else:
            obj = copy.copy(original)
            for k, v in (self.remove or {}).iteritems():
                obj.pop(k)
            for k, v in (self.update or {}).iteritems():
                # Recursive application
                obj[k] = v.apply(obj[k])
            for k, v in (self.append or {}).iteritems():
                if hasattr(obj, 'insert'):
                    obj.insert(k, v)
                else:
                    obj[k] = v
            return obj


class Diff(DiffWrapper):
    """A class to encapsulate differences between two JSON git objects.

    :param obj1: The original object.
    :type obj1: :class:`JsonGitObject`
    :param obj2: The object to compare to.
    :type obj2: :class:`JsonGitObject`
    """

    APPEND = '_append'
    REMOVE = '_remove'
    UPDATE = '_update'

    @classmethod
    def is_json_diff(cls, obj):
        """Determine whether a dict was produced by JSON diff.
        """
        if isinstance(obj, dict):
            return any(k in obj for k in [cls.APPEND, cls.REMOVE, cls.UPDATE])
        else:
            return False

    def __init__(self, obj1, obj2):
        if isinstance(obj2, obj1.__class__):
            c = json_diff.Comparator()
            c.obj1 = obj1
            c.obj2 = obj2
            diff = c._compare_elements(obj1, obj2)
            super(Diff, self).__init__(diff)
        else:
            # if types differ we just replace
            super(Diff, self).__init__(obj2)


class Conflict(object):
    """A class wrapper for the conflict between two diffs.
    """

    def __init__(self, diff1, diff2):
        self._conflict = {}
        if diff1.replace or diff2.replace:
            if diff1.replace != diff2.replace:
                self._conflict = {'replace': (diff1.replace, diff2.replace)}
        else:
            for verb1, verb2 in itertools.product(['append', 'update', 'remove'],
                                                    repeat=2):
                mod1 = getattr(diff1, verb1) or {}
                mod2 = getattr(diff2, verb2) or {}

                # Isolate simultaneously modified keys
                for k in (k for k in mod1 if k in mod2):
                    self._conflict.setdefault(verb1, {})
                    # If verbs were the same, it's OK unless mod was different.
                    if verb1 == verb2 and mod1[k] != mod2[k]:
                        self._conflict[verb1][k] = (mod1[k], mod2[k])
                    # Otherwise, it's a conflict no matter what
                    else:
                        self._conflict[verb1][k] = (mod1[k], None)
                        self._conflict.setdefault(verb2, {})
                        self._conflict[verb2][k] = (None, mod2[k])

    def __nonzero__(self):
        return len(self._conflict) != 0

    def __str__(self):
        return self._conflict.__str__()

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, self._conflict.__repr__())

    @property
    def remove(self):
        """A dict of key removal conflict tuples.
        """
        return self._conflict.get('remove')

    @property
    def update(self):
        """A dict of key update conflict tuples.
        """
        return self._conflict.get('update')

    @property
    def append(self):
        """A dict of key append conflict tuples.
        """
        return self._conflict.get('append')

    @property
    def replace(self):
        """A tuple of the two diffs.
        """
        return self._conflict.get('replace')


class Merge(object):
    """A class wrapper for the results of a merge operation.
    """

    def __init__(self, success, original, merged, message, result=None,
                 conflict=None):
        self._success = success
        self._message = message
        self._original = original
        self._merged = merged
        self._conflict = conflict
        self._result = result

    def __str__(self):
        return self.message

    def __repr__(self):
        return "%s(success=%s,message=%s,conflict=%s,original=%s,merged=%s)" % (
            type(self).__name__, self.success, self.message, self.conflict,
            self.original, self.merged)

    def __nonzero__(self):
        return self.success

    @property
    def result(self):
        """
        :returns:
            The :class:`Object` resulting from this merge, or None if
            there was a conflict.
        :rtype: :class:`Object`
        """
        return self._result

    @property
    def success(self):
        return self._success

    @property
    def original(self):
        return self._original

    @property
    def merged(self):
        return self._merged

    @property
    def conflict(self):
        return self._conflict

    @property
    def message(self):
        return self._message



from helpers import RepoTestCase

class TestObject(RepoTestCase):

    def test_dirty(self):
        """Starts clean, then gets dirty.
        """
        dict = self.repo.commit('key', {})
        self.assertFalse(dict.dirty)
        dict['roses'] = 'red'
        self.assertTrue(dict.dirty)
        dict.commit()
        self.assertFalse(dict.dirty)

    def test_autocommit_never_dirty(self):
        """Can't get dirty if autocommit.
        """
        dict = self.repo.commit('key', {}, autocommit=True)
        dict['roses'] = 'red'
        self.assertFalse(dict.dirty)

    def test_commit(self):
        """Values shouldn't take until commit.
        """
        dict = self.repo.commit('key', {})
        dict['roses'] = 'red'
        self.assertEqual({}, self.repo.get('key').value)
        dict.commit()
        self.assertEqual({'roses': 'red'}, self.repo.get('key').value)

    def test_autocommit(self):
        """Values take immediately if autocommit.
        """
        dict = self.repo.commit('key', {}, autocommit=True)
        dict['roses'] = 'red'
        self.assertEqual({'roses': 'red'}, self.repo.get('key').value)

    def test_nonshared_merge(self):
        """Cannot merge if no shared parent commit.
        """
        foo = self.repo.commit('foo', {'roses': 'red'})
        bar = self.repo.commit('bar', {'violets': 'blue'})
        self.assertFalse(foo.merge(bar).success)
        self.assertFalse(bar.merge(foo).success)

    def test_simple_merge(self):
        """If there's a shared parent and no intervening commits, should
        merge.
        """
        foo = self.repo.commit('foo', {'roses': 'red'})
        bar = self.repo.merge('bar', 'foo').result
        foo['violets'] = 'blue'
        foo.commit()
        merge = bar.merge(foo)
        self.assertTrue(merge.success)
        self.assertEqual({'roses': 'red', 'violets': 'blue'}, bar.value)

    def test_merge_conflict(self):
        """
        If there was a conflict that cannot be automatically resolved, should
        not merge.
        """
        foo = self.repo.commit('foo', {'roses': 'red'})
        bar = self.repo.merge('bar', 'foo').result
        foo['roses'] = 'pink'
        foo.commit()
        bar['roses'] = 'orange'
        bar.commit()
        self.assertFalse(foo.merge(bar).success)
        self.assertFalse(bar.merge(foo).success)

    def test_automatic_merge(self):
        """
        If the changes don't conflict, merges should be automatic.
        """
        foo = self.repo.commit('foo', {'roses': 'red'})
        bar = self.repo.merge('bar', 'foo').result

        foo['violets'] = 'blue'
        foo.commit()

        bar['lilacs'] = 'purple'
        bar.commit()

        merge = foo.merge(bar)
        self.assertTrue(merge.success)
        self.assertEqual({'roses':'red',
                          'violets':'blue',
                          'lilacs': 'purple'}, foo.value)



# -*- coding: utf-8 -*-

"""
jsongit.constants
"""

import pygit2

GIT_SORT_NONE = pygit2.GIT_SORT_NONE
GIT_SORT_TOPOLOGICAL = pygit2.GIT_SORT_TOPOLOGICAL
GIT_SORT_TIME = pygit2.GIT_SORT_TIME
GIT_SORT_REVERSE = pygit2.GIT_SORT_REVERSE



# -*- coding: utf-8 -*-

__title__ = 'jsongit'
__author__ = 'John Krauss'
__license__ = 'BSD'
__copyright__ = 'Copyright 2012 John Krauss'

from .api import init
from .utils import signature, global_config
from .exceptions import NotJsonError, InvalidKeyError, DifferentRepoError, NoGlobalSettingError
from .constants import GIT_SORT_NONE, GIT_SORT_TOPOLOGICAL, GIT_SORT_TIME, GIT_SORT_REVERSE



# -*- coding: utf-8 -*-

"""
jsongit.exceptions
"""

class NotJsonError(ValueError):
    """
    Raised when an object that cannot be run through :func:`json.dumps` is
    committed.  Subclasses :exc:`ValueError`.
    """
    pass


class InvalidKeyError(TypeError):
    """
    Raised when a non-string or invalid string is used as a key in a
    repository.  Subclasses :exc:`TypeError`.
    """
    pass


class DifferentRepoError(ValueError):
    """This is raised if a merge is attempted on a :class:`Object` in a
    different repo.  Subclasses :exc:`ValueError`.
    """
    pass


class NoGlobalSettingError(RuntimeError):
    """Raised when the requested global setting does not exist. Subclasses
    :exc:`RuntimeError`.
    """

    def __init__(self, name):
        super(NoGlobalSettingError, self).__init__(
            "Git on this system has no global setting '%s'" % name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    # Install prereqs here and now if we can.
    from setuptools import setup
    kw = { 'install_requires': [
        'pygit2>=0.16.1',
        'json_diff>=1.2.9'
    ] }
except ImportError:
    from distutils.core import setup
    print 'No setuptools.  Do\n\n    $ pip install pygit2\n    $ pip install json_diff\n\nto install dependencies.'
    kw = {}

execfile('jsongit/version.py')

packages = ['jsongit']

setup(
    name='jsongit',
    version=__version__,
    description='Git for JSON',
    long_description=open('README.rst').read(),
    author='John Krauss',
    author_email='john@accursedware.com',
    url='http://github.com/talos/jsongit',
    packages=packages,
    license='BSD',
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7'
    ),
    **kw
)



# -*- coding: utf-8 -*-

"""
jsongit.models

The meat of jsongit code, particularly the Repository, resides here.
"""

import pygit2
import collections
import functools
import shutil
import itertools

from .exceptions import NotJsonError, InvalidKeyError, DifferentRepoError
from .wrappers import Commit, Diff, Conflict, Merge
import constants
import utils

# The name of the only blob within the special commit tree.

class Repository(object):
    def __init__(self, repo, dumps, loads):
        self._repo = repo
        self._global_name = utils.global_config('user.name')
        self._global_email = utils.global_config('user.email')
        self._dumps = dumps
        self._loads = loads

    def __eq__(self, other):
        return self._repo.path == other._repo.path

    def _key2ref(self, key):
        """The keys of a Repository are also references to the head commit
        for that key.  This translates keys to the appropriate path (in
        /refs/heads/jsongit/).

        :raises: InvalidKeyError
        """
        if not isinstance(key, basestring):
            raise InvalidKeyError("%s must be a string to be a key." % key)
        elif key[-1] == '.' or key[-1] == '/':
            raise InvalidKeyError("Key '%s' should not end in . or /" % key)
        else:
            return 'refs/heads/jsongit/%s' % key

    def _insert_repo_entry(self, key, blob_id, author, committer, message):
        """Insert a new piece of data into the repo's tree.
        """
        # TreeBuilder doesn't support inserting into trees, so we roll our own
        head_target = self._repo.lookup_reference('HEAD').target
        try:
            head_commit = self._repo[self._repo.lookup_reference(head_target).oid]
            head_tree = head_commit.tree
            tree_data = b''.join([
                b"%o %s\x00%s" % (e.attributes, e.name, e.oid)  # octal attributes
                for e in head_tree if e.name != key  # skip existing entry
            ])
            parents = [head_commit.oid]
        except KeyError:  # no commit yet for head (first commit to repo),
                          # so we create an empty tree
            tree_data = ''
            parents = []
        tree_data = tree_data + b"100644 %s\x00%s" % (key, blob_id)

        tree_id = self._repo.write(pygit2.GIT_OBJ_TREE, tree_data)
        self._repo.index.read_tree(tree_id)
        self._repo.index.write()
        # idx = self._repo.index
        # idx.read_tree(tree_id)
        # tree_id = idx.write_tree()
        # idx.clear()
        # idx.read_tree(tree_id)
        # idx.write()
        self._repo.create_commit(head_target, author, committer, message, tree_id, parents)

    def commit(self, key, data, **kwargs):
        """Commit new data to the key.  Maintains relation to parent commits.

        :param key: The key of the new data.
        :type key: string
        :param data:  The value of the item.
        :type data: anything that runs through :func:`json.dumps`
        :param autocommit:
            (optional) whether the retrieved :class:`Object` should autocommit.
            Defaults to False.
        :type autocommit: boolean
        :param author:
            (optional) The signature for the author of the first commit.
            Defaults to global author.
        :param message:
            (optional) Message for first commit.  Defaults to "adding [key]" if
            there was no prior value.
        :type message: string
        :param author:
            (optional) The signature for the committer of the first commit.
            Defaults to global author.
        :type author: pygit2.Signature
        :param committer:
            (optional) The signature for the committer of the first commit.
            Defaults to author.
        :type committer: pygit2.Signature
        :param parents:
            (optional) The parents of this commit.  Defaults to the last commit
            for this key if it already exists, or an empty list if not.
        :type parents: list of :class:`Commit`

        :returns: the committed data, extended with JsonGit methods
        :rtype: :class:`Object`

        :raises: NotJsonError, InvalidKeyError
        """
        message = kwargs.pop('message', '' if self.has(key) else 'adding %s' % key)
        parents = kwargs.pop('parents', [self.get(key).head] if self.has(key) else [])
        author = kwargs.pop('author', utils.signature(self._global_name,
                                                      self._global_email))
        autocommit = kwargs.pop('autocommit', False)
        committer = kwargs.pop('committer', author)
        if kwargs:
            raise TypeError("Unknown keyword args %s" % kwargs)

        try:
            blob_id = self._repo.write(pygit2.GIT_OBJ_BLOB, self._dumps(data))
        except ValueError as e:
            raise NotJsonError(e)
        except TypeError as e:
            raise NotJsonError(e)

        self._insert_repo_entry(key, blob_id, author, committer, message)

        tree_id = self._repo.write(pygit2.GIT_OBJ_TREE, b"100644 %s\x00%s" % (key, blob_id))
        try:
            self._repo.create_commit(self._key2ref(key), author, committer,
                                     message, tree_id,
                                     [parent.oid for parent in parents])
        except pygit2.GitError as e:
            if str(e).startswith('Failed to create reference'):
                raise InvalidKeyError(e)
            else:
                raise e
        else:
            return self.get(key, autocommit=autocommit)

    def destroy(self):
        """Erase this Git repository entirely.  This will remove its directory.
        Methods called on a repository or its objects after it is destroyed
        will throw exceptions.
        """
        shutil.rmtree(self._repo.path)
        self._repo = None

    def has(self, key):
        """Determine whether there is an entry for key in this repository.

        :param key: The key to check
        :type key: string

        :returns: whether there is an entry
        :rtype: boolean

        :raises: InvalidKeyError, KeyError
        """
        try:
            self._repo.lookup_reference(self._key2ref(key))
            return True
        except KeyError:
            return False

    def get(self, key=None, back=0, commit=None, autocommit=False):
        """Obtain the :class:`Object` associated with a key or commit.

        :param key: (optional) The key to look up.
        :type key: string
        :param back:
            (optional) How many steps back from head to get the object.
            Defaults to 0 (the most recent).
        :type back: integer
        :param commit:
            (optional) An explicit commit to get the object for.
        :type commit: :class:`Commit`
        :param autocommit:
            (optional) Whether the retrieved data should commit when changed.
            Defaults to false.
        :type autocommit: boolean

        :returns: the wrapped data
        :rtype: :class:`Object`
        :raises: KeyError if there is no entry for key
        """
        if commit is None:
            try:
                commit = itertools.islice(self.log(key), back, back + 1).next()
            except StopIteration:
                raise IndexError("%s has fewer than %s commits" % (key, back))

        raw_data = self._repo[self._repo[commit.oid].tree[commit.key].oid].data
        return Object(self, commit, self._loads(raw_data), autocommit)

    def merge(self, dest, key=None, commit=None, **kwargs):
        """Try to merge two keys together.  Will create a new value at dest if
        it does not exist, whose first commit will point to the merge source.

        :param dest: the key to receive the merge
        :type dest: string
        :param key:
            (optional) the key of the merge source, which will use the head
            commit.
        :type key: string
        :param commit: (optional) the explicit commit to merge
        :type commit: :class:`Commit`
        :param author:
            (optional) The author of this commit, if one is necessary.
            Defaults to global author.
        :type author: pygit2.Signature
        :param committer:
            (optional) The committer of this commit, if one is necessary.
            Will default to global author.
        :type committer: pygit2.Signature

        :returns: The results of the merge operation
        :rtype: :class:`JsonMerge`
        """
        if commit is None:
            commit = self.get(key).head
        if commit.key == dest:
            raise ValueError('Cannot merge a key with itself')

        # copying
        if not self.has(dest):
            message = "Copying %s from %s" % (dest, commit.key)
            obj = self.commit(dest, commit.object.value, message=message,
                              parents=[commit])
            return Merge(True, commit, obj.head, message, result=obj)

        dest_head = self.get(dest).head
        # No difference
        if commit.oid == dest_head.oid:
            return Merge(True, commit, dest_head, "Same commit", result=commit.object)

        # Do a merge if there were no overlapping changes
        # First, find the shared parent
        try:
            shared_commit = (dc for dc in self.log(commit=dest_head)
                             if dc.oid in (sc.oid for sc in self.log(commit=commit))).next()
        except StopIteration:
            return Merge(False, commit, dest_head, "No shared parent")

        # Now, see if the diffs conflict
        source_diff = Diff(shared_commit.object.value, commit.object.value)
        dest_diff = Diff(shared_commit.object.value, dest_head.object.value)

        conflict = Conflict(source_diff, dest_diff)

        # No-go, the user's gonna have to figure this one out
        if conflict:
            return Merge(False, commit, dest_head, "Merge conflict",
                                conflict=conflict)
        # Sweet. we can apply all the diffs.
        else:
            merged_data = dest_diff.apply(source_diff.apply(shared_commit.object.value))
            message = "Auto-merge from %s" % shared_commit.hex
            parents = [dest_head, commit]
            obj = self.commit(dest, merged_data, message=message, parents=parents, **kwargs)
            return Merge(True, commit, dest_head, message, result=obj)

    def log(self, key=None, commit=None, order=constants.GIT_SORT_TOPOLOGICAL):
        """ Traverse commits from the specified key or commit.  Must specify
        one or the other.

        :param key:
            (optional) The key to look up a log for.  Will look from the head
            commit.
        :type key: string
        :param commit:
            (optional) An explicit commit to look up log for.
        :type commit: :class:`Commit`
        :param order:
            (optional) Flags to order traversal.  Valid flags are in
            :mod:`constants`.  Defaults to :const:`GIT_SORT_TOPOLOGICAL`
        :type order: number

        :returns:
            A generator to traverse commits, yielding :class:`Commit` objects.
        :rtype: generator
        """
        if commit is None:
            c = self._repo[self._repo.lookup_reference(self._key2ref(key)).oid]
            commit = Commit(self, c)
        return (Commit(self, c) for c in self._repo.walk(commit.oid, order))


class Object(collections.MutableMapping, collections.MutableSequence):

    def dirtify(meth):
        """Decorator that dirties up the object upon successful completion.
        """
        @functools.wraps(meth)
        def wrapped(self, *args, **kwargs):
            retval = meth(self, *args, **kwargs)
            self._dirty = True  # if above call fails, we're not dirtied.
            if self.autocommit:
                self.commit()
            return retval
        return wrapped

    def __init__(self, repo, commit, value, autocommit):

        #: Whether changes should be automatically committed.
        self.autocommit = autocommit

        self._repo = repo
        self._head = commit
        self._value = value
        self._dirty = False

    def _value_meth(self, meth):
        cls = self.value.__class__
        try:
            return getattr(cls, meth)
        except AttributeError:
            raise TypeError("%s does not support %s" % (cls, meth))

    def __contains__(self, item):
        return self._value_meth('__contains__')(self.value, item)

    def __len__(self):
        return self._value_meth('__len__')(self.value)

    def __iter__(self):
        return self._value_meth('__iter__')(self.value)

    def __getitem__(self, key):
        return self._value_meth('__getitem__')(self.value, key)

    @dirtify
    def __setitem__(self, key, value):
        return self._value_meth('__setitem__')(self.value, key, value)

    @dirtify
    def __delitem__(self, key):
        return self._value_meth('__delitem__')(self.value, key)

    @dirtify
    def insert(self, item):
        return self._value_meth('insert')(self.value, item)

    def __str__(self):
        return self._value

    def __repr__(self):
        return '%s(key=%s,value=%s,dirty=%s)' % (
            type(self).__name__,
            self.key,
            self._value.__repr__(),
            self.dirty)

    def _update(self, commit=None):
        if commit is None:
            commit = self._repo.get(self.key).head
        self._value = commit.object.value
        self._head = commit
        self._dirty = False

    @property
    def repo(self):
        """The :class:`Repository` of this object.
        """
        return self._repo

    @property
    def key(self):
        """The String key for this dict in its repository.
        """
        return self.head.key

    @property
    def dirty(self):
        """Whether the current value is different than what's in the repo.
        """
        return self._dirty

    @property
    def value(self):
        """The current (possibly dirty) value of this object.
        """
        return self._value

    @property
    def head(self):
        """The :class:`Commit` last associated with this object.
        """
        return self._head

    def commit(self, **kwargs):
        """Convenience wrapper for :func:`Repository.commit` applying to this
        key.  Resets the dirty flag and updates head.
        """
        commit = self.repo.commit(self.key, self.value, **kwargs).head
        self._update(commit=commit)
        assert self.value == commit.object.value

    def log(self, **kwargs):
        """Convenience wrapper for :func:`Repository.log` applying to this key.
        Returns the log based off this object's head, which may not be the
        most recent commit for the key in the repository.
        """
        return self.repo.log(commit=self.head, **kwargs)

    def merge(self, other, **kwargs):
        """Convenience wrapper for :func:`Repository.commit`

        :param other:
            the object to merge in.  The merge is done to the head of
            this object.
        :type other: :class:`Object`

        :raises: DifferentRepoError
        """
        if other.repo == self.repo:
            merge = self.repo.merge(self.key, commit=other.head, **kwargs)
            if merge.success:
                self._update()
            return merge
        else:
            # this would run, but would break references in future.
            raise DifferentRepoError("Cannot merge object in, it's in a \
                                     different repo")
